# -*- coding: utf-8 -*-
"""
State Model Abstract
====================
Modified: 2021-12

Dependencies:
-------------
```
```
"""

import os
import json
from json import JSONDecodeError
from typing import Any, Dict

from myosin.exceptions.cache import CachePathError, NullCachePathError
from myosin.typing import _PKey
from myosin.models.base import BaseModel

BP_ENV_VAR = "MYOSIN_CACHE_BASE_PATH"


class StateModel(BaseModel):

    def __init__(self, _id: _PKey) -> None:
        super().__init__(_id)
        self.cache_base_path = os.environ.get(BP_ENV_VAR)
        self._cpath = f'{self.cache_base_path}/{self.__class__.__name__}.json'

    def cache(self) -> None:
        """
        Serialize contents and save to cache as a json file

        :raises NullCachePathError: _description_
        :raises CachePathError: _description_
        """
        if not self.cache_base_path:
            raise NullCachePathError(
                f"Caching basepath is unset. set the {BP_ENV_VAR} environment variable before using model caching")
        if not os.path.exists(self.cache_base_path):
            raise CachePathError(f"Caching base path {self.cache_base_path} does not exist")
        with open(self._cpath, 'w+') as json_file:
            cached_payload = self.serialize()
            json.dump(cached_payload, json_file)
        self._logger.debug("Cached state model: %s", self)

    def load(self) -> None:
        """
        Load contents from json into state model
        """
        try:
            with open(self._cpath, 'r') as json_file:
                device_payload: Dict[str, Any] = json.load(json_file)
        except JSONDecodeError as exc:
            self._logger.error("Model cache document corrupt:\n%s", exc)
        except FileNotFoundError:
            self._logger.warning("Model not found in caching directory")
        else:
            self.deserialize(**device_payload)
            self._logger.debug("Loaded state model: %s", self)



# -*- coding: utf-8 -*-
"""
State Manager
=============
Modified: 2021-05
Dependancies
------------
```
import logging
import asyncio
from monitor.models.icb import ICB
from monitor.models.device import Device
from monitor.models.protocol import Protocol
from monitor.models.experiment import Experiment
from monitor.models.imaging_profile import ImagingProfile
```
"""
import copy
import logging
import asyncio
import traceback
from typing import Any, Coroutine, Dict, Tuple, Type, TypeVar, Callable, List
from myosin.utils.funcs import pformat
from myosin.utils.concurrency import ThreadUtils as tutils
from myosin.exceptions.state import NullCheckoutError
from myosin.models.state import StateModel

# generic runtime model type
_T = TypeVar('_T', bound=StateModel)


class State:

    # storage mechanism is { name: Model }
    _ssm: Dict[int, Any] = {}

    def __init__(self) -> None:
        self._logger = logging.getLogger(__name__)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb): ...

    def load(self, model: _T) -> _T:
        """
        Load state model into memory

        :param model: state model
        :type model: StateModel
        """
        self._ssm[model.__typehash__()] = model
        self._logger.info("Loaded state model: %s", pformat(model.serialize()))
        return model

    @tutils.lock(tutils.state_lock)
    def checkout(self, _type: Type[_T]) -> _T:
        """
        Returns requested state model

        :return: deep copy of requested state model
        :rtype: StateModel
        """
        self._logger.info("Checking out state model of type %s", _type)
        _type_hash = hash(_type)
        self._logger.debug("Computed type hash: %s", _type_hash)
        mdl = self._ssm.get(_type_hash)
        if not mdl:
            raise NullCheckoutError
        return copy.deepcopy(mdl)

    @tutils.lock(tutils.state_lock)
    def commit(self, state: _T, cache: bool = False) -> _T:
        _type_hash = hash(type(state))
        self._ssm[_type_hash] = state
        if cache:
            state.cache()
            self._logger.debug("Cached commited state model %s", state)
        return state

    def subscribe(self, state_type: Type[_T], callback: Callable[[_T], Coroutine[Any, Any, None]]) -> None:
        """
        Subscribe an asynchronous state change listener to a designated runtime model

        :param state_type: runtime type to subscribe to
        :type state_type: Type[T]
        :param callback: state change listener callback
        :type callback: Callable[[T], Coroutine[Any, Any, None]]
        """

    async def _subscriber_runner(self, state_var: _T,
                                 registry: List[Callable[[_T], Coroutine[Any, Any, None]]]) -> None:
        """
        Async runner that executes all subscriber coroutines with new runtime model state changes.

        :param state_var: pre-validated runtime model to pass to subscribers
        :type state_var: T
        :param registry: list of subscriber coroutines
        :type registry: List[Callable[[T], Awaitable[None]]]
        """
        # construct coroutine lists
        tasks = [subscriber(state_var) for subscriber in registry]
        # return results from coroutines with exceptions if any
        results = await asyncio.gather(*tasks, return_exceptions=True)
        # reformat tasks to display only the name
        operations = tuple(zip(map(lambda x: x.__qualname__, tasks), results))
        self._logger.debug("%s Subscriber operations: %s", type(state_var).__name__, operations)
        # filter by operations which yielded an exception
        exceptions: List[Tuple[str, Exception]] = list(
            filter(lambda x: x[1] is not None, operations))
        for func, exc in exceptions:
            self._logger.exception("Subscriber function: %s encountered an exception: %s", func,
                                   "".join(traceback.format_exception(
                                       etype=type(exc), value=exc, tb=exc.__traceback__
                                   )))



# -*- coding: utf-8 -*-
"""
Concurrency Utils
=================
Modified: 2022-03
"""

from threading import Lock
from typing import Callable, TypeVar

_T = TypeVar('_T')


class ThreadUtils:

    state_lock = Lock()

    @classmethod
    def lock(cls, lock: Lock) -> Callable:
        """
        Apply a semaphore lock to decorated function

        :param lock: lock for resource
        :type lock: Lock
        :return: decorator
        :rtype: Callable
        """
        def decorator(func: Callable[..., _T]) -> Callable[..., _T]:
            def wrapper(*args, **kwargs) -> _T:
                with lock:
                    response = func(*args, **kwargs)
                return response
            return wrapper
        return decorator



# -*- coding: utf-8 -*-
"""
Base Model Abstract
===================
Modified: 2022-03
"""

import logging
from typing import Any, Dict
from abc import ABC, abstractmethod

from myosin.typing import _PKey
from myosin.utils.funcs import pformat


class BaseModel(ABC):

    def __init__(self, _id: _PKey) -> None:
        self._logger = logging.getLogger(__name__)
        self.id = _id

    def __typehash__(self) -> int:
        return hash(type(self))

    def __eq__(self, o: object) -> bool:
        if hasattr(o, 'id') and hasattr(self, 'id'):
            return o.id == self.id  # type: ignore
        return False

    def __repr__(self) -> str:
        return "{}".format(pformat(self.serialize()))

    @property
    def id(self) -> _PKey:
        """
        Get state model id

        :return: state model id
        :rtype: _PKey
        """
        return self.__id

    @id.setter
    def id(self, _id: _PKey) -> None:
        """
        Set state model id

        :param _id: state model id 
        :type _id: _PKey
        """
        self.__id = _id

    @abstractmethod
    def serialize(self) -> Dict[str, Any]: ...

    @abstractmethod
    def deserialize(self, **kwargs) -> None: ...
